/*------------------------------------------------------------------------------*
 * Architecture & Implementation of DBMS                                        *
 *------------------------------------------------------------------------------*
 * Copyright 2019 Databases and Information Systems Group TU Dortmund          *
 *                                                                              *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS      *
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,  *
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL      *
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR         *
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,        *
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR        *
 * OTHER DEALINGS IN THE SOFTWARE.                                              *
 *------------------------------------------------------------------------------*
 */
#ifndef simple_tables_include
#define simple_tables_include
#pragma once

#include <vector>
#include <array>
#include <memory>
#include <cassert>
#include <iostream>
#include <string>
#include <functional>

#include <algorithm>


namespace aidbms::simple_tables {


    using VALUE_TYPE = int32_t;

//    bool EQ (const VALUE_TYPE left, const VALUE_TYPE right) { return left == right; }

    struct Predicate {
        std::string left;
        std::string right;
    };

    template <typename CONTAINER, typename VALUE>
    class iterator : public std::iterator< std::input_iterator_tag, VALUE, std::size_t, const VALUE*, VALUE> {
    public:
        iterator(std::size_t i, const CONTAINER& t):_i(i),_t(t) {
        }
        const iterator& operator++() {
            _i++;
            return *this;
        }
        bool operator!=(iterator other) const {
            return _i != other._i;
        }
        VALUE operator*() const {
            return _t[_i]; // works when CONTAINER implements operator[] that returns VALUE
        }
    private:
        std::size_t _i;
        const CONTAINER& _t;
    };


    /**
     * @brief The Tuple struct is a simple wrapper class for a pointer into const (!) values of a relation.
     * Gives convenient (read-only) access to tuple values via iterator.
     */
    struct Tuple {
        Tuple(const VALUE_TYPE* tuple_addr, std::size_t M): _tuple_addr(tuple_addr), _M(M) {}
        Tuple(int x) : _tuple_addr(0), _M(0){ }
        Tuple() : _tuple_addr(0), _M(0) { }

        friend std::ostream& operator<<(std::ostream& from, const Tuple& t) {
            for (auto value: t) {
                from << value;
            }
            return from;
        }
        VALUE_TYPE operator[](std::size_t column_idx) const {
            return _tuple_addr[column_idx];
        }
        operator bool() const {
            return _tuple_addr != 0;
        }
        std::size_t size() const {
            return _M;
        }
        iterator<Tuple, VALUE_TYPE> begin() const {
            return iterator<Tuple, VALUE_TYPE>(0,*this);
        }
        iterator<Tuple, VALUE_TYPE> end() const {
            return iterator<Tuple, VALUE_TYPE>(_M,*this);
        }

    private:
        const VALUE_TYPE* _tuple_addr;
        std::size_t _M;

    };

    class Relation {
    public:
        /**
         * @brief Relation constructor that is usually used to create an empty relation with no tuples
         * and fixed dimensionality. Data can only be added and removed tuple-wise via "addTuple" and
         * "removeTuple" respectively.
         * @param column_names names of the columns. number of names also gives dimensionality!
         * @param relation_name name of this relation
         */
        Relation (std::vector<std::string> column_names, std::string relation_name):
            _M(column_names.size()),_N(0),_column_names(column_names),_relation_name(relation_name) {

        }

        Relation (const Relation &r1, const Relation &r2) {
            std::vector<std::string> t_names;
            t_names.insert(t_names.end(), r1.columnNames().begin(), r1.columnNames().end());
            t_names.insert(t_names.end(), r2.columnNames().begin(), r2.columnNames().end());

            _M = t_names.size();
            _N = 0;
            _relation_name = r1.name() + std::string("x") + r2.name();
        }
        /**
         * @brief addTuple Adds a tuple at the bottom of the relation. Can trigger relocation and data movement.
         * Make use of member function reserve(), if the relation's final size is known in advance!
         * @param tuple values in an array, that are copied into the relation.
         */
        void addTuple(std::vector<VALUE_TYPE>&& tuple) {
            assert(_M == tuple.size()); /* We require the user to add tuple of fixed length M! */
            // TODO: call _values.reserve?
            for (auto&& value: tuple) {
                _values.push_back(value);
            }
            _N++;
        }
        /**
         * @brief addTuple for two given (smaller) tuples. their length has to add up to _M of this relation!
         * @param r_tuple
         * @param s_tuple
         */
        void addTuple(Tuple r_tuple, Tuple s_tuple) {
            assert(_M == (r_tuple.size()+s_tuple.size())); /* dim. check */
            // TODO: call _values.reserve?
            for (auto value: r_tuple) {
                _values.push_back(value);
            }
            for (auto value: s_tuple) {
                _values.push_back(value);
            }
            _N++;
        }
        /**
         * @brief removeTuple deletes the selected tuple. Since Relation is based on a vector of its values,
         * deletions can trigger data movement in order to keep the data in a contiguous chunk!
         *
         * @param tuple_idx A valid tuple index. Guarded by an assertion!
         */
        void removeTuple(std::size_t tuple_idx) {
            assert((tuple_idx >= 0) && (tuple_idx < _N));
            // erase all values of the tuple at once:
            _values.erase(_values.begin()+(tuple_idx*_M), _values.begin()+((tuple_idx+1)*_M));
            _N--; // decrease row count
        }


        ////////////////////////////// getters: ////////////////////////////////////////////////
        /**
         * @brief numTuples is a imple getter.
         * @return number of tuples
         */
        std::size_t size() const {
            return _N;
        }
        /**
         * @brief M gives the number of columns
         * @return
         */
        std::size_t M() const {
            return _M;
        }
        /**
         * @brief columnNames gives the column names in regular order.
         * @return
         */
        const std::vector<std::string>& columnNames() const {
            return _column_names;
        }

        std::size_t columnIndex(std::string column_name) const  {
            auto column = std::find(_column_names.begin(),_column_names.end(),column_name);
            if (column != _column_names.end()) { // we found the column!
                return static_cast<std::size_t>(column-_column_names.begin());
            }
            else {
                throw std::invalid_argument(std::string("Column not found in ")+ _relation_name);
            }
        }

        /**
         * @brief name gives the name of the relation
         * @return
         */
        const std::string& name() const {
            return _relation_name;
        }

        /**
         * @brief operator [] gives (unchecked!) access to tuples.
         * This access can not be used to alter the relation!
         * @param row is the index of the requested tuple
         * @return A pointer to the first value of the tuple, encapsulated in "Tuple"
         */
        Tuple operator[](std::size_t tuple_idx) const {
            return Tuple((&_values[0])+tuple_idx*_M,_M); // unchecked access to the contiguous chunk of values
        }


        ////////////////////////////// utils: ///////////////////////////////////////////////////
        /**
         * @brief reserveTuples If the (new) size of the relation is known in advance, memory can be allocated in a
         * single step (in "sizeof tuple"-steps).
         * @param num_tuples The new size of the relation in the number of tuples.
         */
        void reserveTuples(std::size_t num_tuples) {
            _values.reserve(num_tuples * _M);
        }

        /**
         * @brief print dumps the complete relation into the specified output stream.
         * @param outstream
         * @param delimiter
         */
        void print(std::ostream& outstream = std::cout, char delimiter = '\t') const {
            outstream << _relation_name << ":" << std::endl;
            for (auto&& name: _column_names)
                outstream << name << delimiter;
            outstream << std::endl;
            for ( std::size_t i = 1; i <= _values.size(); i++) {
                outstream << _values[i-1];
                if (i % _M)
                    outstream << delimiter;
                else // if this is the last value in a tuple: add a newline
                    outstream << '\n';
            }
        }

        /**
         * @brief begin for ranged based loop
         * @return iterator pointing to the first Tuple
         */
        iterator<Relation, Tuple> begin() const {
            return iterator<Relation, Tuple>(0,*this);
        }
        /**
         * @brief end for ranged based loop
         * @return iterator pointing to the end of the relation
         */
        iterator<Relation, Tuple> end() const {
            return iterator<Relation, Tuple>(_N,*this);
        }
    private:
        std::size_t _M; // relation's dimension/number of attributes
        std::size_t _N; // keep track of the relations number of tuples with an internal counter
//        bool locked; // if true, we do not allow to alter this relation
        std::vector<std::string> _column_names;
        std::string _relation_name;
        std::vector<VALUE_TYPE> _values; // the actual data in a contiguous chunk of memory aka. a vector
    };
} // namespace end

#endif
